#include <iostream>
#ifdef __MINGW32__
#include <tr1/unordered_map>
using namespace std::tr1;
#else
#include <unordered_map>
#endif

#include <map>
#include <fstream>
#include <vector>
#include <string>
#include <map>
#include <cmath>
#include <cstring>
#include <string>
#include "Utility.h"
#include "colored-gdv.h"

int ntype2gtype[] = {0,  1, 1, 2, 3, 3, 4, 4, 5, 6, 6, 6, 7, 7, 8, 9, 9, 9, 10, 10,
10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 15, 16, 16, 16, 16,
17, 17, 17, 17, 18, 18, 19, 19, 19, 19, 20, 20, 21, 21, 21, 22, 22, 23, 23,
23, 24, 24, 24, 25, 25, 25, 26, 26, 26, 27, 27, 28, 28, 29};

inline int GetEdgeColorIndex(vector<int> nodes, map<pair<int,int>, int> &edgeColors, char **adjMatrix) {
    int result = 0;
    for (auto it1 = nodes.begin(); it1 != nodes.end(); ++it1) {
        for (auto it2 = it1+1; it2 != nodes.end(); ++it2) {
            if (AreNodesConnected(*it1, *it2)) {
                result |= edgeColors[{*it1, *it2}];
            }
        }
    }

    return (result >> 1) - 1;
}

enum {
    NODE,
    EDGE
};

void compute_gdv(int argc, char **argv) {
    if(argc < 2)
        Exit("Usage: "+ToString(argv[0])+"[-node, -edge] [input-network: GW format] [input-color-file] [output-node-vectors]");
    int graphletType;
    if (!strcmp(argv[1], "-node") && argc != 4) {
        graphletType = NODE; 
    } else if (!strcmp(argv[1], "-edge") && argc != 3) {
        graphletType = EDGE;
    } else {
        Exit("Usage: "+ToString(argv[0])+"[-node, -edge] [input-network: GW format] [input-color-file] [output-node-vectors]");
    }
    
    ifstream fiNetwork, fiColor;
    fiNetwork.open(argv[2]);
    if(!fiNetwork.is_open())
        Exit("Error: File "+ToString(argv[2])+" not found");

    if(graphletType == NODE) {
        fiColor.open(argv[3]);
        if(!fiColor.is_open())
            Exit("Error: File "+ToString(argv[3])+" not found");
    }

    int currentProcess=0;
    int totalProcess=1;
    for(int i=graphletType == NODE ? 5 : 4;i<argc;i+=2) {
        string paramType(argv[i]);
        if(i+1 >= argc)
            Exit("Error: Missing value for "+paramType);
        string paramVal(argv[i+1]);

        if(paramType == "-i")
            currentProcess = ToInteger(paramVal);
        else if(paramType == "-p")
            totalProcess = ToInteger(paramVal);
        else
            Exit("Error: Invalid parameter type "+paramType);
    }
    if(currentProcess >= totalProcess)
        Exit("current-process should be less than total-process");

    string line;
    fiNetwork >> line;
    fiNetwork >> line;
    fiNetwork >> line;
    fiNetwork >> line;
    int noNodes;
    if(IsPositiveInteger(line)) {
        noNodes = ToInteger(line);
    }
    else {
        fiNetwork >> line;
        if(IsPositiveInteger(line))
            noNodes = ToInteger(line);
        else
            Exit("Error: Incorrect data format in "+ToString(argv[2]));
    }

    int mask[MAX_COLORS-1];
    mask[0] = 1;
    for(int i=1;i<MAX_COLORS;i++) {
        mask[i] = mask[i-1]*2;
    }

    Node *nodeList = new Node[noNodes];
    map<pair<int,int>, int> edgeColors; // use struct Edge because pairs are commutative
    map<string,int> nodeHash;
    char **adjMatrix = new char*[noNodes];
    for(int i=0;i<noNodes;i++) {
        fiNetwork >> line;
        if(line.substr(0,2) != "|{")
            Exit("Error: Incorrect data format in "+ToString(argv[2]));
        if(line.substr(line.length()-2,2) != "}|")
            Exit("Error: Incorrect data format in "+ToString(argv[2]));
        nodeList[i].name = line.substr(2,line.length()-4);
        nodeHash[nodeList[i].name] = i;
        nodeList[i].color = 0;
        adjMatrix[i] = (char *) calloc(noNodes/8+4,sizeof(char));
        if(!adjMatrix[i]) Exit("Error: Bad memory allocation by calloc");
    }

    fiNetwork >> line;
    int noEdges;
    if(IsPositiveInteger(line))
        noEdges = ToInteger(line);
    else
        Exit("Error: Incorrect data format in "+ToString(argv[1]));

    int node1,node2,edgeColor;
    string dummy1,dummy2;
    int noColors = 0;
    map<string,int>::iterator it;
    for(int i=0;i<noEdges;i++) {
        fiNetwork >> node1 >> node2 >> edgeColor >> dummy1;
        node1--;
        node2--;
        if(node1 == node2)
            continue;
        if(node1<0 || node2<0)
            Exit("Error: Incorrect data format in "+ToString(argv[1]));

        if (!AreNodesConnected(node1,node2)) {
            nodeList[node1].link.push_back(node2);
            nodeList[node2].link.push_back(node1);
            ConnectNodes(node1,node2);
            ConnectNodes(node2,node1);
        }
        if (graphletType == EDGE) { // Compute edge colors during input file
            // Add color to mask for both node1 and node2
            if(node1 < noNodes && node2 < noNodes) {
                if (edgeColors.find({node1,node2}) != edgeColors.end()) {
                    edgeColors[{node1,node2}] |= mask[edgeColor];
                    edgeColors[{node2,node1}] |= mask[edgeColor];
                } else {
                    edgeColors[{node1,node2}] = mask[edgeColor];
                    edgeColors[{node2,node1}] = mask[edgeColor];
                }
                if (edgeColor > noColors) {
                    noColors = edgeColor;
                }
            }
        }
    }
    fiNetwork.close();

    string nodeName;
    int colorVal;
    int noNodesWithColor = 0;
    if (graphletType == NODE) { // Compute node colors reading separate color file
        while(fiColor >> nodeName >> colorVal) {
            it = nodeHash.find(nodeName);
            if(it != nodeHash.end() && nodeList[it->second].color == 0) {
                nodeList[it->second].color = mask[colorVal];
                noNodesWithColor++;
                if(colorVal > noColors)
                    noColors = colorVal;
            }
        }
        if(noNodes != noNodesWithColor) {
            Exit("Not all the nodes in "+ToString(argv[1])+" are colored by "+ToString(argv[2]));
        }
        fiColor.close();
    }
    int noColorCombinations = pow(2,noColors) - 1;

    if (noColors > MAX_COLORS) {
        Exit("Maximum number of colors is " + ToString(MAX_COLORS));
    }

    for(int i=0;i<noNodes;i++) {
        nodeList[i].ncount.resize(NO_ORBITS*noColorCombinations,0);
    }

    int colorIndex;

    for(int i=0;i<noNodes;i++) {
        int a = i;

        int noLinkA = nodeList[a].link.size();
        for(int j=0;j<noLinkA;j++) {
            int b = nodeList[a].link[j];
            if(b==a)
                continue;
            if (graphletType == NODE) {
                colorIndex = ((nodeList[a].color | nodeList[b].color) >> 1) - 1;
            } else {
                colorIndex = GetEdgeColorIndex({a,b}, edgeColors, adjMatrix);
            }
            nodeList[a].ncount[LINK*noColorCombinations + colorIndex]++;

            int noLinkB = nodeList[b].link.size();
            for(int k=0;k<noLinkB;k++) {
                int c = nodeList[b].link[k];
                if(c==a || c==b)
                    continue;
                if (graphletType == NODE) {
                    colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color) >> 1) - 1;
                } else {
                    colorIndex = GetEdgeColorIndex({a,b,c}, edgeColors, adjMatrix);
                }
                int deg3_A=0,deg3_B=0,deg3_C=0;
                if(AreNodesConnected(a,b)) { deg3_A++; deg3_B++; }
                if(AreNodesConnected(a,c)) { deg3_A++; deg3_C++; }
                if(AreNodesConnected(b,c)) { deg3_B++; deg3_C++; }

                if(deg3_A == 1) {
                    nodeList[a].ncount[P3_A*noColorCombinations + colorIndex]++;
                    nodeList[b].ncount[P3_B*noColorCombinations + colorIndex]++;
                    nodeList[c].ncount[P3_A*noColorCombinations + colorIndex]++;

                    if(nodeList[b].link.size() > 2) {
                        for(int l=0;l<noLinkB;l++) {
                            int d = nodeList[b].link[l];
                            if(d==a || d==b || d==c)
                                continue;
                            if (graphletType == NODE) {
                                colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color) >> 1) - 1;
                            } else {
                                colorIndex = GetEdgeColorIndex({a,b,c,d}, edgeColors, adjMatrix);
                            }
                            if(!AreNodesConnected(a,d) && !AreNodesConnected(c,d)) {
                                nodeList[a].ncount[CLAW_A*noColorCombinations + colorIndex]++;
                                nodeList[c].ncount[CLAW_A*noColorCombinations + colorIndex]++;
                                nodeList[d].ncount[CLAW_A*noColorCombinations + colorIndex]++;
                                nodeList[b].ncount[CLAW_B*noColorCombinations + colorIndex]++;

                                for(int m=0;m<noLinkB;m++) {
                                    int e = nodeList[b].link[m];
                                    if(e==a || e==b || e==c || e==d)
                                        continue;
                                    if (graphletType == NODE) {
                                        colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color | nodeList[e].color) >> 1) - 1;
                                    } else {
                                        colorIndex = GetEdgeColorIndex({a,b,c,d,e}, edgeColors, adjMatrix);
                                    }
                                    if(!AreNodesConnected(a,e) && !AreNodesConnected(c,e) && !AreNodesConnected(d,e)) {
                                        nodeList[a].ncount[X11_A*noColorCombinations + colorIndex]++;
                                        nodeList[b].ncount[X11_B*noColorCombinations + colorIndex]++;
                                        nodeList[c].ncount[X11_A*noColorCombinations + colorIndex]++;
                                        nodeList[d].ncount[X11_A*noColorCombinations + colorIndex]++;
                                        nodeList[e].ncount[X11_A*noColorCombinations + colorIndex]++;
                                    }
                                }
                            }
                        }
                    }
                }
                else {
                    nodeList[a].ncount[C3_A*noColorCombinations + colorIndex]++;
                    nodeList[b].ncount[C3_A*noColorCombinations + colorIndex]++;
                    nodeList[c].ncount[C3_A*noColorCombinations + colorIndex]++;
                }

                int noLinkC = nodeList[c].link.size();
                for(int l=0;l<noLinkC;l++) {
                    int d = nodeList[c].link[l];

                    if(d==a || d==b || d==c)
                        continue;
                    if (graphletType == NODE) {
                        colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color) >> 1) - 1;
                    } else {
                        colorIndex = GetEdgeColorIndex({a,b,c,d}, edgeColors, adjMatrix);
                    }

                    int deg4_A = deg3_A, deg4_B = deg3_B, deg4_C = deg3_C, deg4_D = 0;
                    if(AreNodesConnected(a,d)) { deg4_A++; deg4_D++; };
                    if(AreNodesConnected(b,d)) { deg4_B++; deg4_D++; };
                    if(AreNodesConnected(c,d)) { deg4_C++; deg4_D++; };
                    int noEdges4 = deg4_A + deg4_B + deg4_C + deg4_D;

                    switch(noEdges4) {
                        case 6:
                            nodeList[a].ncount[P4_A*noColorCombinations + colorIndex]++;
                            nodeList[b].ncount[P4_B*noColorCombinations + colorIndex]++;
                            nodeList[c].ncount[P4_B*noColorCombinations + colorIndex]++;
                            nodeList[d].ncount[P4_A*noColorCombinations + colorIndex]++;

                            for(int m=0;m<noLinkB;m++) {
                                int e = nodeList[b].link[m];
                                if(e==a || e==b || e==c || e==d)
                                    continue;
                                if (graphletType == NODE) {
                                    colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color | nodeList[e].color) >> 1) - 1;
                                } else {
                                    colorIndex = GetEdgeColorIndex({a,b,c,d,e}, edgeColors, adjMatrix);
                                }
                                if(!AreNodesConnected(a,e) && !AreNodesConnected(c,e) && !AreNodesConnected(d,e)) {
                                    nodeList[a].ncount[X10_B*noColorCombinations + colorIndex]++;
                                    nodeList[b].ncount[X10_D*noColorCombinations + colorIndex]++;
                                    nodeList[c].ncount[X10_C*noColorCombinations + colorIndex]++;
                                    nodeList[d].ncount[X10_A*noColorCombinations + colorIndex]++;
                                    nodeList[e].ncount[X10_B*noColorCombinations + colorIndex]++;
                                }
                            }
                            break;
                        case 8:
                            if(deg4_B == 3 || deg4_C == 3) {
                                if(deg4_B == 3) {
                                    nodeList[a].ncount[FLOW_A*noColorCombinations + colorIndex]++;
                                    nodeList[b].ncount[FLOW_C*noColorCombinations + colorIndex]++;
                                    nodeList[c].ncount[FLOW_B*noColorCombinations + colorIndex]++;
                                    nodeList[d].ncount[FLOW_B*noColorCombinations + colorIndex]++;

                                    for(int m=0;m<noLinkB;m++) {
                                        int e = nodeList[b].link[m];
                                        if(e==a || e==b || e==c || e==d)
                                            continue;
                                        if (graphletType == NODE) {
                                            colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color | nodeList[e].color) >> 1) - 1;
                                        } else {
                                            colorIndex = GetEdgeColorIndex({a,b,c,d,e}, edgeColors, adjMatrix);
                                        }
                                        if(!AreNodesConnected(a,e) && !AreNodesConnected(c,e) && !AreNodesConnected(d,e)) {
                                            nodeList[a].ncount[X14_A*noColorCombinations + colorIndex]++;
                                            nodeList[b].ncount[X14_C*noColorCombinations + colorIndex]++;
                                            nodeList[c].ncount[X14_B*noColorCombinations + colorIndex]++;
                                            nodeList[d].ncount[X14_B*noColorCombinations + colorIndex]++;
                                            nodeList[e].ncount[X14_A*noColorCombinations + colorIndex]++;
                                        }
                                    }
                                }
                                else {
                                    nodeList[a].ncount[FLOW_B*noColorCombinations + colorIndex]++;
                                    nodeList[b].ncount[FLOW_B*noColorCombinations + colorIndex]++;
                                    nodeList[c].ncount[FLOW_C*noColorCombinations + colorIndex]++;
                                    nodeList[d].ncount[FLOW_A*noColorCombinations + colorIndex]++;
                                }
                            }
                            else {
                                nodeList[a].ncount[C4_A*noColorCombinations + colorIndex]++;
                                nodeList[b].ncount[C4_A*noColorCombinations + colorIndex]++;
                                nodeList[c].ncount[C4_A*noColorCombinations + colorIndex]++;
                                nodeList[d].ncount[C4_A*noColorCombinations + colorIndex]++;
                            }
                            break;
                        case 10:
                            nodeList[a].ncount[deg4_A==3 ? DIAM_B*noColorCombinations + colorIndex : DIAM_A*noColorCombinations + colorIndex]++;
                            nodeList[b].ncount[deg4_B==3 ? DIAM_B*noColorCombinations + colorIndex : DIAM_A*noColorCombinations + colorIndex]++;
                            nodeList[c].ncount[deg4_C==3 ? DIAM_B*noColorCombinations + colorIndex : DIAM_A*noColorCombinations + colorIndex]++;
                            nodeList[d].ncount[deg4_D==3 ? DIAM_B*noColorCombinations + colorIndex : DIAM_A*noColorCombinations + colorIndex]++;
                            break;
                        case 12:
                            nodeList[a].ncount[K4_A*noColorCombinations + colorIndex]++;
                            nodeList[b].ncount[K4_A*noColorCombinations + colorIndex]++;
                            nodeList[c].ncount[K4_A*noColorCombinations + colorIndex]++;
                            nodeList[d].ncount[K4_A*noColorCombinations + colorIndex]++;
                            break;
                    }

                    int noLinkD = nodeList[d].link.size();
                    for(int m=0;m<noLinkD;m++) {
                        int e = nodeList[d].link[m];
                        if(e==a || e==b || e==c || e==d)
                            continue;
                        if (graphletType == NODE) {
                            colorIndex = ((nodeList[a].color | nodeList[b].color | nodeList[c].color | nodeList[d].color | nodeList[e].color) >> 1) - 1;
                        } else {
                            colorIndex = GetEdgeColorIndex({a,b,c,d,e}, edgeColors, adjMatrix);
                        }
                        int deg5_A = deg4_A, deg5_B = deg4_B, deg5_C = deg4_C, deg5_D = deg4_D, deg5_E = 0;
                        if(AreNodesConnected(a,e)) { deg5_A++; deg5_E++; };
                        if(AreNodesConnected(b,e)) { deg5_B++; deg5_E++; };
                        if(AreNodesConnected(c,e)) { deg5_C++; deg5_E++; };
                        if(AreNodesConnected(d,e)) { deg5_D++; deg5_E++; };

                        int nDegA = deg5_A, nDegB = deg5_B, nDegC = deg5_C, nDegD = deg5_D, nDegE = deg5_E;
                        if(AreNodesConnected(a,b)) { nDegA += deg5_B; nDegB += deg5_A; };
                        if(AreNodesConnected(a,c)) { nDegA += deg5_C; nDegC += deg5_A; };
                        if(AreNodesConnected(a,d)) { nDegA += deg5_D; nDegD += deg5_A; };
                        if(AreNodesConnected(a,e)) { nDegA += deg5_E; nDegE += deg5_A; };
                        if(AreNodesConnected(b,c)) { nDegB += deg5_C; nDegC += deg5_B; };
                        if(AreNodesConnected(b,d)) { nDegB += deg5_D; nDegD += deg5_B; };
                        if(AreNodesConnected(b,e)) { nDegB += deg5_E; nDegE += deg5_B; };
                        if(AreNodesConnected(c,d)) { nDegC += deg5_D; nDegD += deg5_C; };
                        if(AreNodesConnected(c,e)) { nDegC += deg5_E; nDegE += deg5_C; };
                        if(AreNodesConnected(d,e)) { nDegD += deg5_E; nDegE += deg5_D; };

                        int hash = nDegA%4 + nDegB%4 + nDegC%4 + nDegD%4 + nDegE%4;
                        int degTotal = deg5_A + deg5_B + deg5_C + deg5_D + deg5_E;
                        int gtype = gtable[degTotal/2 - 4][hash/2];

                        if(degTotal == 14 && hash == 6) {
                            gtype = (nDegA > 12 || nDegA == 5) ? 22 : 24;
                        }
                        int indexA = (int)ntable[gtype][nDegA]+1;
                        int indexB = (int)ntable[gtype][nDegB]+1;
                        int indexC = (int)ntable[gtype][nDegC]+1;
                        int indexD = (int)ntable[gtype][nDegD]+1;
                        int indexE = (int)ntable[gtype][nDegE]+1;
                        nodeList[a].ncount[indexA*noColorCombinations + colorIndex]++;
                        nodeList[b].ncount[indexB*noColorCombinations + colorIndex]++;
                        nodeList[c].ncount[indexC*noColorCombinations + colorIndex]++;
                        nodeList[d].ncount[indexD*noColorCombinations + colorIndex]++;
                        nodeList[e].ncount[indexE*noColorCombinations + colorIndex]++;
                    }
                }
            }
        }
    }

    ofstream fo(graphletType == NODE ? argv[4] : argv[3]);
    //ofstream fo("output.txt");
    for(int i=0;i<noNodes;i++) {
        fo << nodeList[i].name;
        for(int j=0;j<NO_ORBITS;j++) {
            for(int k=0;k<noColorCombinations;k++) {
                fo << "\t" << nodeList[i].ncount[j*noColorCombinations + k]/overcount[ntype2gtype[j]];
            }
        }
        fo << endl;
    }
    fo.close();

    return;
}
